@@ -1,6 +1,6 @@
-#include <byebug.h>
+#include "byebug.h"
 
-static VALUE mByebug;   /* Ruby Byebug Module object */
+static VALUE mByebug; /* Ruby Byebug Module object */
 
 static VALUE tracing = Qfalse;
 static VALUE post_mortem = Qfalse;
@@ -13,6 +13,7 @@ static VALUE tracepoints = Qnil;
 static VALUE raised_exception = Qnil;
 
 static ID idPuts;
+static ID idEmpty;
 
 /* Hash table with active threads and their associated contexts */
 VALUE threads = Qnil;
@@ -36,9 +37,9 @@ Breakpoints(VALUE self)
 
 /*
  *  call-seq:
- *    Byebug.catchpoints -> array
+ *    Byebug.catchpoints -> hash
  *
- *  Returns an array of catchpoints.
+ *  Returns the catchpoints hash.
  */
 static VALUE
 Catchpoints(VALUE self)
@@ -62,7 +63,8 @@ Raised_exception(VALUE self)
   return raised_exception;
 }
 
-#define IS_STARTED  (catchpoints != Qnil)
+#define IS_STARTED (!NIL_P(catchpoints))
+
 static void
 check_started()
 {
@@ -73,7 +75,7 @@ check_started()
 }
 
 static void
-trace_print(rb_trace_arg_t * trace_arg, debug_context_t * dc,
+trace_print(rb_trace_arg_t *trace_arg, debug_context_t *dc,
             const char *file_filter, const char *debug_msg)
 {
   char *fullpath = NULL;
@@ -125,7 +127,7 @@ trace_print(rb_trace_arg_t * trace_arg, debug_context_t * dc,
 }
 
 static void
-cleanup(debug_context_t * dc)
+cleanup(debug_context_t *dc)
 {
   dc->stop_reason = CTX_STOP_NONE;
 
@@ -147,30 +149,42 @@ cleanup(debug_context_t * dc)
   thread_context_lookup(rb_thread_current(), &context); \
   Data_Get_Struct(context, debug_context_t, dc);        \
                                                         \
-  if (CTX_FL_TEST(dc, CTX_FL_IGNORE))                   \
-    return;                                             \
-                                                        \
-  acquire_lock(dc);                                     \
-                                                        \
   trace_arg = rb_tracearg_from_tracepoint(trace_point); \
   if (verbose == Qtrue)                                 \
     trace_print(trace_arg, dc, 0, 0);                   \
+                                                        \
+  if (CTX_FL_TEST(dc, CTX_FL_IGNORE))                   \
+    return;                                             \
+                                                        \
+  acquire_lock(dc);
+
+
+#define CALL_EVENT_SETUP   \
+  dc->calced_stack_size++; \
+  dc->steps_out = dc->steps_out < 0 ? -1 : dc->steps_out + 1;
+
+#define RETURN_EVENT_SETUP \
+  dc->calced_stack_size--; \
+                           \
+  if (dc->steps_out == 1)  \
+    dc->steps = 1;
+
+#define RETURN_EVENT_TEARDOWN \
+  dc->steps_out = dc->steps_out <= 0 ? -1 : dc->steps_out - 1;
 
 
 /* Functions that return control to byebug after the different events */
 
 static VALUE
-call_at(VALUE context_obj, debug_context_t * dc, ID mid, int argc, VALUE a0,
-        VALUE a1)
+call_at(VALUE ctx, debug_context_t *dc, ID mid, int argc, VALUE arg)
 {
   struct call_with_inspection_data cwi;
-  VALUE argv[2];
+  VALUE argv[1];
 
-  argv[0] = a0;
-  argv[1] = a1;
+  argv[0] = arg;
 
   cwi.dc = dc;
-  cwi.context_obj = context_obj;
+  cwi.ctx = ctx;
   cwi.id = mid;
   cwi.argc = argc;
   cwi.argv = &argv[0];
@@ -179,49 +193,60 @@ call_at(VALUE context_obj, debug_context_t * dc, ID mid, int argc, VALUE a0,
 }
 
 static VALUE
-call_at_line(VALUE context_obj, debug_context_t * dc, VALUE file, VALUE line)
+call_at_line(VALUE ctx, debug_context_t *dc)
 {
-  return call_at(context_obj, dc, rb_intern("at_line"), 2, file, line);
+  return call_at(ctx, dc, rb_intern("at_line"), 0, Qnil);
 }
 
 static VALUE
-call_at_tracing(VALUE context_obj, debug_context_t * dc, VALUE file, VALUE line)
+call_at_tracing(VALUE ctx, debug_context_t *dc)
 {
-  return call_at(context_obj, dc, rb_intern("at_tracing"), 2, file, line);
+  return call_at(ctx, dc, rb_intern("at_tracing"), 0, Qnil);
 }
 
 static VALUE
-call_at_breakpoint(VALUE context_obj, debug_context_t * dc, VALUE breakpoint)
+call_at_breakpoint(VALUE ctx, debug_context_t *dc, VALUE breakpoint)
 {
   dc->stop_reason = CTX_STOP_BREAKPOINT;
-  return call_at(context_obj, dc, rb_intern("at_breakpoint"), 1, breakpoint, 0);
+
+  return call_at(ctx, dc, rb_intern("at_breakpoint"), 1, breakpoint);
 }
 
 static VALUE
-call_at_catchpoint(VALUE context_obj, debug_context_t * dc, VALUE exp)
+call_at_catchpoint(VALUE ctx, debug_context_t *dc, VALUE exp)
 {
   dc->stop_reason = CTX_STOP_CATCHPOINT;
-  return call_at(context_obj, dc, rb_intern("at_catchpoint"), 1, exp, 0);
+
+  return call_at(ctx, dc, rb_intern("at_catchpoint"), 1, exp);
 }
 
 static VALUE
-call_at_return(VALUE context_obj, debug_context_t * dc, VALUE file, VALUE line)
+call_at_return(VALUE ctx, debug_context_t *dc, VALUE return_value)
 {
   dc->stop_reason = CTX_STOP_BREAKPOINT;
-  return call_at(context_obj, dc, rb_intern("at_return"), 2, file, line);
+
+  return call_at(ctx, dc, rb_intern("at_return"), 1, return_value);
+}
+
+static VALUE
+call_at_end(VALUE ctx, debug_context_t *dc)
+{
+  dc->stop_reason = CTX_STOP_BREAKPOINT;
+
+  return call_at(ctx, dc, rb_intern("at_end"), 0, Qnil);
 }
 
 static void
-call_at_line_check(VALUE context_obj, debug_context_t * dc, VALUE breakpoint,
-                   VALUE file, VALUE line)
+call_at_line_check(VALUE ctx, debug_context_t *dc, VALUE breakpoint)
 {
   dc->stop_reason = CTX_STOP_STEP;
 
-  if (breakpoint != Qnil)
-    call_at_breakpoint(context_obj, dc, breakpoint);
+  if (!NIL_P(breakpoint))
+    call_at_breakpoint(ctx, dc, breakpoint);
 
   reset_stepping_stop_points(dc);
-  call_at_line(context_obj, dc, file, line);
+
+  call_at_line(ctx, dc);
 }
 
 
@@ -239,7 +264,7 @@ line_event(VALUE trace_point, void *data)
   binding = rb_tracearg_binding(trace_arg);
 
   if (RTEST(tracing))
-    call_at_tracing(context, dc, file, line);
+    call_at_tracing(context, dc);
 
   if (!CTX_FL_TEST(dc, CTX_FL_IGNORE_STEPS))
     dc->steps = dc->steps <= 0 ? -1 : dc->steps - 1;
@@ -253,15 +278,17 @@ line_event(VALUE trace_point, void *data)
   }
 
   if (dc->steps == 0 || dc->lines == 0)
-    call_at_line_check(context, dc, Qnil, file, line);
+    call_at_line_check(context, dc, Qnil);
+  else
+  {
+    brkpnt = Qnil;
 
-  brkpnt = Qnil;
+    if (!NIL_P(breakpoints))
+      brkpnt = find_breakpoint_by_pos(breakpoints, file, line, binding);
 
-  if (!NIL_P(breakpoints))
-    brkpnt = find_breakpoint_by_pos(breakpoints, file, line, binding);
-
-  if (!NIL_P(brkpnt))
-    call_at_line_check(context, dc, brkpnt, file, line);
+    if (!NIL_P(brkpnt))
+      call_at_line_check(context, dc, brkpnt);
+  }
 
   EVENT_TEARDOWN;
 }
@@ -269,16 +296,14 @@ line_event(VALUE trace_point, void *data)
 static void
 call_event(VALUE trace_point, void *data)
 {
-  VALUE brkpnt, klass, msym, mid, binding, self, file, line;
+  VALUE brkpnt, klass, msym, mid, binding, self;
 
   EVENT_SETUP;
 
   if (dc->calced_stack_size <= dc->dest_frame)
     CTX_FL_UNSET(dc, CTX_FL_IGNORE_STEPS);
 
-  dc->calced_stack_size++;
-
-  dc->steps_out = dc->steps_out <= 0 ? -1 : dc->steps_out + 1;
+  CALL_EVENT_SETUP;
 
   /* nil method_id means we are at top level so there can't be a method
    * breakpoint here. Just leave then. */
@@ -293,8 +318,6 @@ call_event(VALUE trace_point, void *data)
   klass = rb_tracearg_defined_class(trace_arg);
   binding = rb_tracearg_binding(trace_arg);
   self = rb_tracearg_self(trace_arg);
-  file = rb_tracearg_path(trace_arg);
-  line = rb_tracearg_lineno(trace_arg);
 
   brkpnt = Qnil;
 
@@ -304,7 +327,7 @@ call_event(VALUE trace_point, void *data)
   if (!NIL_P(brkpnt))
   {
     call_at_breakpoint(context, dc, brkpnt);
-    call_at_line(context, dc, file, line);
+    call_at_line(context, dc);
   }
 
   EVENT_TEARDOWN;
@@ -313,23 +336,69 @@ call_event(VALUE trace_point, void *data)
 static void
 return_event(VALUE trace_point, void *data)
 {
+  VALUE brkpnt, file, line, binding;
+
   EVENT_SETUP;
 
-  dc->calced_stack_size--;
+  RETURN_EVENT_SETUP;
 
-  if (dc->steps_out == 1)
-    dc->steps = 1;
-  else if ((dc->steps_out == 0) && (CTX_FL_TEST(dc, CTX_FL_STOP_ON_RET)))
+  if ((dc->steps_out == 0) && (CTX_FL_TEST(dc, CTX_FL_STOP_ON_RET)))
   {
-    VALUE file, line;
-
     reset_stepping_stop_points(dc);
+
+    call_at_return(context, dc, rb_tracearg_return_value(trace_arg));
+  }
+  else if (!NIL_P(breakpoints))
+  {
     file = rb_tracearg_path(trace_arg);
-    line = rb_tracearg_lineno(trace_arg);
-    call_at_return(context, dc, file, line);
+    /*
+     * @todo Sometimes the TracePoint API gives some return events without
+     * file:line information, so we need to guard for nil until we know what's
+     * going on. This happens, for example, with active_support core extensions:
+     *
+     * [#7] call@.../core_ext/numeric/conversions.rb:124 Fixnum#to_s
+     *  [#7] b_call@.../core_ext/numeric/conversions.rb:124 BigDecimal#to_s
+     *   [#7] line@.../core_ext/numeric/conversions.rb:125 BigDecimal#to_s
+     *   [#7] c_call@.../core_ext/numeric/conversions.rb:125 Kernel#is_a?
+     *    [#7] c_return@.../core_ext/numeric/conversions.rb:125 Kernel#is_a?
+     *   [#7] line@.../core_ext/numeric/conversions.rb:131 BigDecimal#to_s
+     *   [#7] c_call@.../core_ext/numeric/conversions.rb:131 Fixnum#to_default_s
+     *    [#7] c_return@.../core_ext/numeric/conversions.rb:131 Fixnum#to_default_s
+     *   [#7] b_return@/hort/core_ext/numeric/conversions.rb:133 BigDecimal#to_s
+     *  [#7] return@:0 Fixnum#to_s # => This guy...
+     */
+    if (!NIL_P(file))
+    {
+      line = rb_tracearg_lineno(trace_arg);
+      binding = rb_tracearg_binding(trace_arg);
+
+      brkpnt = find_breakpoint_by_pos(breakpoints, file, line, binding);
+
+      if (!NIL_P(brkpnt))
+        call_at_return(context, dc, rb_tracearg_return_value(trace_arg));
+    }
   }
 
-  dc->steps_out = dc->steps_out <= 0 ? -1 : dc->steps_out - 1;
+  RETURN_EVENT_TEARDOWN;
+
+  EVENT_TEARDOWN;
+}
+
+static void
+end_event(VALUE trace_point, void *data)
+{
+  EVENT_SETUP;
+
+  RETURN_EVENT_SETUP;
+
+  if ((dc->steps_out == 0) && (CTX_FL_TEST(dc, CTX_FL_STOP_ON_RET)))
+  {
+    reset_stepping_stop_points(dc);
+
+    call_at_end(context, dc);
+  }
+
+  RETURN_EVENT_TEARDOWN;
 
   EVENT_TEARDOWN;
 }
@@ -339,7 +408,7 @@ raw_call_event(VALUE trace_point, void *data)
 {
   EVENT_SETUP;
 
-  dc->calced_stack_size++;
+  CALL_EVENT_SETUP;
 
   EVENT_TEARDOWN;
 }
@@ -349,7 +418,9 @@ raw_return_event(VALUE trace_point, void *data)
 {
   EVENT_SETUP;
 
-  dc->calced_stack_size--;
+  RETURN_EVENT_SETUP;
+
+  RETURN_EVENT_TEARDOWN;
 
   EVENT_TEARDOWN;
 }
@@ -357,32 +428,24 @@ raw_return_event(VALUE trace_point, void *data)
 static void
 raise_event(VALUE trace_point, void *data)
 {
-  VALUE expn_class, ancestors;
-  VALUE path, lineno, binding, post_mortem_context;
+  VALUE expn_class, ancestors, pm_context;
   int i;
   debug_context_t *new_dc;
 
   EVENT_SETUP;
 
-  path = rb_tracearg_path(trace_arg);
-  lineno = rb_tracearg_lineno(trace_arg);
-  binding = rb_tracearg_binding(trace_arg);
   raised_exception = rb_tracearg_raised_exception(trace_arg);
 
   if (post_mortem == Qtrue)
   {
-    post_mortem_context = context_dup(dc);
-    rb_ivar_set(raised_exception, rb_intern("@__bb_file"), path);
-    rb_ivar_set(raised_exception, rb_intern("@__bb_line"), lineno);
-    rb_ivar_set(raised_exception, rb_intern("@__bb_binding"), binding);
-    rb_ivar_set(raised_exception, rb_intern("@__bb_context"),
-                post_mortem_context);
+    pm_context = context_dup(dc);
+    rb_ivar_set(raised_exception, rb_intern("@__bb_context"), pm_context);
 
-    Data_Get_Struct(post_mortem_context, debug_context_t, new_dc);
+    Data_Get_Struct(pm_context, debug_context_t, new_dc);
     rb_debug_inspector_open(context_backtrace_set, (void *)new_dc);
   }
 
-  if (catchpoints == Qnil || dc->calced_stack_size == 0
+  if (NIL_P(catchpoints) || dc->calced_stack_size == 0
       || RHASH_TBL(catchpoints)->num_entries == 0)
   {
     EVENT_TEARDOWN;
@@ -400,11 +463,13 @@ raise_event(VALUE trace_point, void *data)
     hit_count = rb_hash_aref(catchpoints, module_name);
 
     /* increment exception */
-    if (hit_count != Qnil)
+    if (!NIL_P(hit_count))
     {
       rb_hash_aset(catchpoints, module_name, INT2FIX(FIX2INT(hit_count) + 1));
+
       call_at_catchpoint(context, dc, raised_exception);
-      call_at_line(context, dc, path, lineno);
+      call_at_line(context, dc);
+
       break;
     }
   }
@@ -427,7 +492,8 @@ register_tracepoints(VALUE self)
   {
     int line_msk = RUBY_EVENT_LINE;
     int call_msk = RUBY_EVENT_CALL;
-    int ret_msk = RUBY_EVENT_RETURN | RUBY_EVENT_B_RETURN | RUBY_EVENT_END;
+    int ret_msk = RUBY_EVENT_RETURN | RUBY_EVENT_B_RETURN;
+    int end_msk = RUBY_EVENT_END;
     int raw_call_msk = RUBY_EVENT_C_CALL | RUBY_EVENT_B_CALL | RUBY_EVENT_CLASS;
     int raw_ret_msk = RUBY_EVENT_C_RETURN;
     int raise_msk = RUBY_EVENT_RAISE;
@@ -435,6 +501,7 @@ register_tracepoints(VALUE self)
     VALUE tpLine = rb_tracepoint_new(Qnil, line_msk, line_event, 0);
     VALUE tpCall = rb_tracepoint_new(Qnil, call_msk, call_event, 0);
     VALUE tpReturn = rb_tracepoint_new(Qnil, ret_msk, return_event, 0);
+    VALUE tpEnd = rb_tracepoint_new(Qnil, end_msk, end_event, 0);
     VALUE tpCCall = rb_tracepoint_new(Qnil, raw_call_msk, raw_call_event, 0);
     VALUE tpCReturn = rb_tracepoint_new(Qnil, raw_ret_msk, raw_return_event, 0);
     VALUE tpRaise = rb_tracepoint_new(Qnil, raise_msk, raise_event, 0);
@@ -443,6 +510,7 @@ register_tracepoints(VALUE self)
     rb_ary_push(traces, tpLine);
     rb_ary_push(traces, tpCall);
     rb_ary_push(traces, tpReturn);
+    rb_ary_push(traces, tpEnd);
     rb_ary_push(traces, tpCCall);
     rb_ary_push(traces, tpCReturn);
     rb_ary_push(traces, tpRaise);
@@ -546,8 +614,6 @@ Current_context(VALUE self)
 
   UNUSED(self);
 
-  check_started();
-
   thread_context_lookup(rb_thread_current(), &context);
 
   return context;
@@ -564,7 +630,7 @@ Started(VALUE self)
 {
   UNUSED(self);
 
-  return IS_STARTED;
+  return IS_STARTED ? Qtrue : Qfalse;
 }
 
 /*
@@ -585,7 +651,6 @@ Stop(VALUE self)
 
     breakpoints = Qnil;
     catchpoints = Qnil;
-    threads = Qnil;
 
     return Qfalse;
   }
@@ -593,6 +658,39 @@ Stop(VALUE self)
   return Qtrue;
 }
 
+static VALUE
+Stoppable(VALUE self)
+{
+  VALUE context;
+  debug_context_t *dc;
+
+  if (!IS_STARTED)
+    return Qfalse;
+
+  if (!NIL_P(breakpoints) && rb_funcall(breakpoints, idEmpty, 0) == Qfalse)
+    return Qfalse;
+
+  if (!NIL_P(catchpoints) && rb_funcall(catchpoints, idEmpty, 0) == Qfalse)
+    return Qfalse;
+
+  if (post_mortem == Qtrue)
+    return Qfalse;
+
+  if (RTEST(tracing))
+    return Qfalse;
+
+  context = Current_context(self);
+  if (!NIL_P(context))
+  {
+    Data_Get_Struct(context, debug_context_t, dc);
+
+    if (dc->steps > 0)
+      return Qfalse;
+  }
+
+  return Qtrue;
+}
+
 /*
  *  call-seq:
  *    Byebug.start -> bool
@@ -623,7 +721,7 @@ Start(VALUE self)
  *  +stop+ parameter forces byebug to stop at the first line of code in +file+
  */
 static VALUE
-Debug_load(int argc, VALUE * argv, VALUE self)
+Debug_load(int argc, VALUE *argv, VALUE self)
 {
   VALUE file, stop, context;
   debug_context_t *dc;
@@ -655,36 +753,6 @@ Debug_load(int argc, VALUE * argv, VALUE self)
   return status;
 }
 
-/*
- *  call-seq:
- *    Byebug.verbose? -> bool
- *
- *  Returns +true+ if verbose output of TracePoint API events is enabled.
- */
-static VALUE
-Verbose(VALUE self)
-{
-  UNUSED(self);
-
-  return verbose;
-}
-
-/*
- *  call-seq:
- *    Byebug.verbose = bool
- *
- *  Enable verbose output of every TracePoint API events, useful for debugging
- *  byebug.
- */
-static VALUE
-Set_verbose(VALUE self, VALUE value)
-{
-  UNUSED(self);
-
-  verbose = RTEST(value) ? Qtrue : Qfalse;
-  return value;
-}
-
 /*
  *  call-seq:
  *    Byebug.tracing? -> bool
@@ -714,6 +782,35 @@ Set_tracing(VALUE self, VALUE value)
   return value;
 }
 
+/*
+ *  call-seq:
+ *    Byebug.verbose? -> bool
+ *
+ *  Returns +true+ if global verbose flag for TracePoint API events is enabled.
+ */
+static VALUE
+Verbose(VALUE self)
+{
+  UNUSED(self);
+
+  return verbose;
+}
+
+/*
+ *  call-seq:
+ *    Byebug.verbose = bool
+ *
+ *  Sets the global verbose flag for TracePoint API events is enabled.
+ */
+static VALUE
+Set_verbose(VALUE self, VALUE value)
+{
+  UNUSED(self);
+
+  verbose = RTEST(value) ? Qtrue : Qfalse;
+  return value;
+}
+
 /*
  *  call-seq:
  *    Byebug.post_mortem? -> bool
@@ -747,7 +844,7 @@ Set_post_mortem(VALUE self, VALUE value)
  *  call-seq:
  *    Byebug.add_catchpoint(exception) -> exception
  *
- *  Adds a new exception to the catchpoints array.
+ *  Adds a new exception to the catchpoints hash.
  */
 static VALUE
 Add_catchpoint(VALUE self, VALUE value)
@@ -786,6 +883,7 @@ Init_byebug()
   rb_define_module_function(mByebug, "start", Start, 0);
   rb_define_module_function(mByebug, "started?", Started, 0);
   rb_define_module_function(mByebug, "stop", Stop, 0);
+  rb_define_module_function(mByebug, "stoppable?", Stoppable, 0);
   rb_define_module_function(mByebug, "thread_context", Thread_context, 1);
   rb_define_module_function(mByebug, "tracing?", Tracing, 0);
   rb_define_module_function(mByebug, "tracing=", Set_tracing, 1);
@@ -803,4 +901,5 @@ Init_byebug()
   rb_global_variable(&threads);
 
   idPuts = rb_intern("puts");
+  idEmpty = rb_intern("empty?");
 }
