@@ -1,5 +1,7 @@
-require 'byebug/setting'
-require 'byebug/commands/list'
+# frozen_string_literal: true
+
+require "byebug/setting"
+require "byebug/commands/list"
 
 module Byebug
   #
@@ -13,7 +15,7 @@ module Byebug
     end
 
     def banner
-      'Invoke list command on every stop'
+      "Invoke list command on every stop"
     end
 
     def value=(v)
